/*

*/

import renderer.scene.*;
import renderer.models.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;
import renderer.gui.*;

import java.awt.Color;
import java.awt.event.*;

/**
   Draw an animation of a solar system with a sun, earth, and moon..
<p>
   In this version, the earth orbits around the sun at the same rate as the sun
   rotates on its axis. Similarly, the moon orbits around the earth at the same
   rate as the earth rotates on its axis.
<p>
   But the orbit of the earth should be independent of the rotation of the sun,
   and the orbit of the moon should be independent of the rotation of the earth.
<pre>
            SolarSystem
                |
                |
             Position
            /   |   \
          /     |     \
     Matrix  Model     nested Positions
       R     (sun)          |
                            |
                         Position
                         /  |  \
                       /    |    \
                 Matrix   Model   nested Positions
                   TR    (earth)       |
                                       |
                                    Position
                                    /  |    \
                                  /    |      \
                              Matrix  Model    nested Positions
                                TR   (moon)        |
                                                   |
                                                 empty
</pre>
*/
public class SolarSystem
{
   private static int fps = 20;

   private static double earthOrbitRadius = 5.0;
   private static double  moonOrbitRadius = 1.0;

   private static double earthOrbitRot = 0.0;
   private static double  moonOrbitRot = 0.0;

   private static double  moonAxisRot = 0.0;

   private static double ecliptic = 7.0; // angle of the ecliptic plane

   private static void print_help_message()
   {
      System.out.println("Use the f/F keys to slow down or speed up the frame rate.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the e/E keys to change the angle of the ecliptic plane.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }

   public static void main(String[] args)
   {
      print_help_message();

      /*
         See the above picture of the tree that this code creates.
      */

      // Create the Scene object that we shall render
      Scene scene = new Scene();  // The Solar System

      // Create the sun.
      Position solarSys = new Position( new Sphere(1.0, 10, 10) );
      solarSys.model.setColor( Color.yellow );

      // Create the planet.
      solarSys.addNestedPosition(new Position( new Sphere(0.5, 10, 10) ));
      solarSys.getNestedPosition(0).model.setColor( Color.blue );

      // Create the planet's moon.
      solarSys.getNestedPosition(0).addNestedPosition(new Position( new Sphere(0.2, 10, 10) ));
      solarSys.getNestedPosition(0).getNestedPosition(0).model.setColor( Color.green );

      // Create references to the nested positions.
      Position earthMoon = solarSys.getNestedPosition(0);
      Position moon = earthMoon.getNestedPosition(0);

      // Add the top level Position to the Scene.
      scene.positionList.add(solarSys);


      // Define initial dimensions for a FrameBuffer.
      int width  = 512;
      int height = 512;

      // Create an AnimationFrame containing a FrameBuffer
      // with the given dimensions.
      @SuppressWarnings("serial")
      InteractiveFrame app = new AnimationFrame("Renderer 8", width, height, fps)
      {
         // Implement the ActionListener interface.
         @Override public void actionPerformed(ActionEvent e)
         {
            //System.out.println( e );

            // Set the model-to-view transformation matrix.
            solarSys.matrix2Identity();
            // Push the model away from where the camera is.
            solarSys.translate(0, 0, -8);
            // Rotate the plane of the ecliptic
            // (rotate the model's plane about the x axis).
            solarSys.rotateX(ecliptic);

            // Set the model matrices for the sub models.
            solarSys.rotateY(earthOrbitRot);

            earthMoon.matrix2Identity();
            earthMoon.translate(earthOrbitRadius, 0, 0);
            earthMoon.rotateY(moonOrbitRot);

            moon.matrix2Identity();
            moon.translate(moonOrbitRadius, 0, 0);
            moon.rotateY(moonAxisRot);

            // Update the parameters for the next frame.
            earthOrbitRot += 1.0;
            moonOrbitRot += 15.0;
            moonAxisRot += 18.0;

            // Render again.
            FrameBuffer fb = this.fbp.getFrameBuffer();
            fb.clearFB(Color.black);
            Pipeline.render(scene, fb);
            fbp.update();
            repaint();
         }

         // Implement part of the KeyListener interface.
         @Override public void keyTyped(KeyEvent e)
         {
            //System.out.println( e );

            char c = e.getKeyChar();
            if ('h' == c)
            {
               print_help_message();
               return;
            }
            else if ('f' == c)
            {
               fps -= 1;
               if (0 == fps) fps = 1;
               System.out.println("fps = " + fps);
            }
            else if ('F' == c)
            {
               fps += 1;
               System.out.println("fps = " + fps);
            }
            else if ('p' == c)
            {
               scene.camera.perspective = ! scene.camera.perspective;
            }
            else if ('e' == c)
            {
               ecliptic -= 1;
               if (0 == fps) fps = 1;
               System.out.println("ecliptic = " + ecliptic);
            }
            else if ('E' == c)
            {
               ecliptic += 1;
               System.out.println("ecliptic = " + ecliptic);
            }

            setFPS(fps);

            // Set up the camera's view volume.
            if (scene.camera.perspective)
            {
               scene.camera.projPerspective(-1, 1, -1, 1, 1);
            }
            else
            {
               scene.camera.projOrtho(-6, 6, -6, 6);
            }
         }

         // Implement part of the ComponentListener interface.
         @Override public void componentResized(ComponentEvent e)
         {
            //System.out.println( e );

            // Get the new size of the FrameBufferPanel.
            int w = this.fbp.getWidth();
            int h = this.fbp.getHeight();

            // Create a new FrameBuffer that fits the new window size.
            FrameBuffer fb = new FrameBuffer(w, h);
            this.fbp.setFrameBuffer(fb);
            fb.clearFB(Color.black);
            Pipeline.render(scene, fb);
            fbp.update();
            repaint();
         }
      };
   }
}
