/*

*/

package renderer.pipeline;
import  renderer.scene.*;
import  renderer.framebuffer.*;

import java.util.List;
import java.util.ArrayList;

/**
   This renderer takes as its input a {@link Scene} data structure
   and a {@link FrameBuffer} data structure. This renderer mutates
   the {@link FrameBuffer}'s current viewport so that it is filled
   in with the rendered image of the scene represented by the
   {@link Scene} object.
<p>
   This implements our eigth rendering pipeline. This renderer
   implements "hierarchical scenes" by recursively traversing the
   DAG of nested {@link Position}s below each of a {@link Scene}'s
   {@link Position} objects. As the renderer traverses deeper into
   the DAG of nested {@link Position}s, it accumulates a "current
   transformation matrix" that transforms a {@link Vertex} from its
   {@link Model}'s local coordinate system to the {@link Camera}'s
   (shared) view coordinate system. The recursion is not a new pipeline
   stage, so there are still six pipeline stages.
*/
public class Pipeline
{
   public static boolean doAntialiasing = false;
   public static boolean debug = false;

   /**
      Mutate the {@link FrameBuffer}'s current viewport so that
      it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(Scene scene, FrameBuffer fb)
   {
      if ( debug )
         render_debug(scene, fb);
      else
         render_(scene, fb);
   }


   /**
      Renderer the {@link Scene} without any debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_(Scene scene, FrameBuffer fb)
   {
      // For every Position in the Scene, render the Position's Model
      // and every nested Position.
      for (Position position : scene.positionList)
      {
         if ( position.visible )
         {
            // Begin a pre-order, depth-first-traversal from this Position.
            render_position(scene, position, Matrix.identity(), fb);
         }
      }
   }


   /**
      Recursively renderer a {@link Position} without any debugging information.
      <p>
      This method does a pre-order, depth-first-traversal of the tree of
      {@link Position}'s rooted at the parameter {@code position}.
      <p>
      The pre-order "visit node" operation in this traversal first updates the
      "current transformation matrix", ({@code ctm}), using the {@link Matrix}
      in {@code position} and then renders the {@link Model} in {@code position}
      using the updated {@code ctm} in the {@link Model2View} stage.


      @param scene     the {@link Scene} that we are rendering
      @param position  the current {@link Position} object to recursively render
      @param ctm       current model-to-view transformation {@link Matrix}
      @param fb        {@link FrameBuffer} to hold rendered image of the {@link Position}
   */
   private static void render_position(Scene scene, Position position, Matrix ctm, FrameBuffer fb)
   {
      // Update the current model-to-view transformation matrix.
      ctm = ctm.times( position.modelMatrix );

      // Render the Position's Model (if it exits and is visible).
      if ( position.model != null && position.model.visible )
      {
         // 0. Make a deep copy of the position's model.
         Model model2 = new Model(position.model);

         // 1. Apply the current model-to-view coordinate transformation.
         Model2View.model2view(model2.vertexList, ctm);

         // 2. Apply the normalizing view-to-camera coordinate transformation.
         View2Camera.view2camera(model2.vertexList, scene.camera);

         // 3. Apply the projection transformation.
         Projection.project(model2.vertexList, scene.camera);

         // 4. Clip each line segment to the standard view rectangle.
         List<LineSegment> lineSegmentList2 = new ArrayList<>();
         for (LineSegment ls : model2.lineSegmentList)
         {
            if ( Clip.clip(ls) )
            {
               // Keep the line segments that are visible.
               lineSegmentList2.add(ls);
            }
         }
         model2.lineSegmentList = lineSegmentList2;

         // 5. Apply viewport transformation to each visible line segment.
         // (Alternatively, we could apply viewport transformation to every vertex?)
         boolean[] transformed = new boolean[model2.vertexList.size()];
         for (LineSegment ls : model2.lineSegmentList)
         {
            Viewport.viewport(ls, transformed, fb);
         }

         // 6. Rasterize each visible line segment into pixels.
         for (LineSegment ls : model2.lineSegmentList)
         {
            if ( doAntialiasing )
            {
               RasterizeAntialias.rasterize(ls, fb);
            }
            else
            {
               Rasterize.rasterize(ls, fb);
            }
         }
      }

      // Recursively render every nested Position of this Position.
      for (Position p : position.nestedPositions)
      {
         if ( p.visible )
         {
            // Do a pre-order, depth-first-traversal from this nested Position.
            render_position(scene, p, ctm, fb);
         }
      }
   }



   /**
      Renderer the {@link Scene} with quite a bit of debugging information.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   private static void render_debug(Scene scene, FrameBuffer fb)
   {
      System.out.println("=== Render Scene ===================================");

      // For every Position in the Scene, render the Position's Model
      // and every nested Position.
      for (Position position : scene.positionList)
      {
         System.out.println("=== Render Position ================================");
         if ( position.visible )
         {
            // Begin a (pre-order) depth-first-traversal from this Position.
            render_position_debug(scene, position, Matrix.identity(), fb);
         }
         else
         {
            System.out.println("   Hidden position.");
         }
         System.out.println("=== End Position ================================");
      }

      System.out.println("=== End Scene ===================================");
   }


   /**
      Recursively renderer a {@link Position} with quite a bit of debugging information.
      <p>
      This method does a pre-order, depth-first-traversal of the tree of
      {@link Position}'s rooted at the parameter {@code position}.
      <p>
      The pre-order "visit node" operation in this traversal first updates the
      "current transformation matrix", ({@code ctm}), using the {@link Matrix}
      in {@code position} and then renders the {@link Model} in {@code position}
      using the updated {@code ctm} in the {@link Model2View} stage.

      @param scene     the {@link Scene} that we are rendering
      @param position  the current {@link Position} object to recursively render
      @param ctm       current model-to-view transformation {@link Matrix}
      @param fb        {@link FrameBuffer} to hold rendered image of the {@link Position}
   */
   private static void render_position_debug(Scene scene, Position position, Matrix ctm, FrameBuffer fb)
   {
      // Update the current model-to-view transformation matrix.
      ctm = ctm.times( position.modelMatrix );

      System.out.println("=== Render Model ===================================");

      // Render the Position's Model (if it exits and is visible).
      if ( position.model != null && position.model.visible )
      {
         // 0. Make a deep copy of the position's model.
         Model model2 = new Model(position.model);

         // log interesting information to standard output
         logVertexList("0 Model  ", model2.vertexList);

         // 1. Apply the current model-to-view coordinate transformation.
         Model2View.model2view(model2.vertexList, ctm);

         // log interesting information to standard output
         logVertexList("1 View   ", model2.vertexList);

         // 2. Apply the normalizing view-to-camera coordinate transformation.
         View2Camera.view2camera(model2.vertexList, scene.camera);

         // log interesting information to standard output
         logVertexList("2 Camera ", model2.vertexList);

         // 3. Apply the projection transformation.
         Projection.project(model2.vertexList, scene.camera);

         // log interesting information to standard output
         logVertexList("3 Project", model2.vertexList);

         // 4. Clip each line segment to the standard view rectangle.
         List<LineSegment> lineSegmentList2 = new ArrayList<>();
         for (LineSegment ls : model2.lineSegmentList)
         {
            // log interesting information to standard output
            logLineSegment("4 Clipping", ls);

            if ( Clip.clip(ls) )
            {
               // log interesting information to standard output
               logLineSegment("4 Clipped (accept)", ls);

               // Keep the line segments that are visible.
               lineSegmentList2.add(ls);
            }
            else
            {
               // log interesting information to standard output
               logLineSegment("4 Clipped (reject)", ls);
            }
         }
         model2.lineSegmentList = lineSegmentList2;

         // log interesting information to standard output
         logVertexList("4 Clipped", model2.vertexList);

         // 5. Apply viewport transformation to each visible line segment.
         boolean[] transformed = new boolean[model2.vertexList.size()];
         for (LineSegment ls : model2.lineSegmentList)
         {
            Viewport.viewport(ls, transformed, fb);

            // log interesting information to standard output
            logLineSegment("5 Viewport", ls);
         }

         // 6. Rasterize each visible line segment into pixels.
         for (LineSegment ls : model2.lineSegmentList)
         {
            // log interesting information to standard output
            logLineSegment("6 Rasterize", ls);

            if ( doAntialiasing )
            {
               RasterizeAntialias.rasterize(ls, fb);
            }
            else
            {
               Rasterize.rasterize(ls, fb);
            }
         }
         System.out.println();
      }
      else
      {
         System.out.println("   Hidden model.");
      }

      System.out.println("=== End Model ===================================");

      // Recursively render every nested Position of this Position.
      for (Position p : position.nestedPositions)
      {
         System.out.println("====== Render Nested Position ========================");
         if ( p.visible )
         {
            // Do a (pre-order) depth-first-traversal from this nested Position.
            render_position_debug(scene, p, ctm, fb);
         }
         else
         {
            System.out.println("   Hidden position.");
         }
         System.out.println("====== End Nested Position ===========================");
      }
   }


   /**
      This method prints a {@link String} representation of the given vertex list.

      @param stage       name for the pipeline stage
      @param vertexList  vertex list whose string representation is to be printed
   */
   private static void logVertexList(String stage, List<Vertex> vertexList)
   {
      int i = 0;
      for (Vertex v : vertexList)
      {
         System.out.printf( stage + ": index=%3d, " + v.toString(4), i );
         i++;
      }
   }


   /**
      This method prints a {@link String} representation of the given {@link LineSegment}.

      @param stage  name for the pipeline stage
      @param ls     {@link LineSegment} whose string representation is to be printed
   */
   private static void logLineSegment(String stage, LineSegment ls)
   {
      System.out.print( stage + ": " + ls.toString() );
      int index0 = ls.index[0];
      int index1 = ls.index[1];
      Vertex v0 = ls.model.vertexList.get(index0);
      Vertex v1 = ls.model.vertexList.get(index1);
      System.out.printf("    index=%3d, " + v0.toString(), index0);
      System.out.printf("               " + v0.toStringColor());
      System.out.printf("    index=%3d, " + v1.toString(), index1);
      System.out.printf("               " + v1.toStringColor(), index1);
   }
}
