/*

*/

package renderer.pipeline;
import  renderer.scene.*;

import java.util.List;

/**
   Transform each {@link Vertex} of a {@link Model} from
   view coordinates to normalized camera coordinates.
<p>
   This stage transforms the {@link Camera}'s view volume from a
   user defined shape (in the view coordinate system) into the
   standard normalized view volume (in the camera coordinate system).
   used by the {@link Clip} pipeline stage.
<p>
   There are two standard normalized view volumes, one for perspective
   projection and one for orthographic projection.
<p>
   The standard normalized perspective view volume is the infinitely
   long pyramid with its apex at the origin and intersecting the plane
   {@code z = -1} at the corners {@code (-1, -1, -1)} and
   {@code (+1, +1, -1)}.
<p>
   The standard normalized orthographic view volume is the infinitely
   long parallelepiped centered on the z-axis and intersecting the
   xy-plane at the corners {@code (-1, -1, 0)} and {@code (+1, +1, 0)}.
<p>
   The user defined view volume (determined by the {@link Scene}'s
   {@link Camera} object) is either the infinitely long pyramid with its
   apex at the origin and intersecting the plane {@code z = -near} at the
   corners {@code (left, bottom, -near)} and {@code (right, top, -near)},
   or it is the infinitely long parallelepiped parallel to the z-axis and
   intersecting the xy-plane at the corners {@code (left, bottom, 0)}
   and {@code (right, top, 0)}.
<p>
   The view coordinate system is relative to the user defined view volume.
<p>
   The normalized camera coordinate system is relative to the normalized
   view volume.
<p>
   The formulas that transform the user defined view volume into the
   normalized view volume also transform the view coordinate system into
   the normalized camera coordinate system.
*/
public class View2Camera_formulas
{
   /**
      Transform each {@link Vertex} from the {@link Camera}'s view
      coordinate system to the normalized camera coordinate system.

      @param vertexList  {@link List} of {@link Vertex} objects to transform into normalized camera coordinates
      @param camera      the {@link Scene}'s {@link Camera} with the view volume data
   */
   public static void view2camera(List<Vertex> vertexList, Camera camera)
   {
      double l = camera.left; // This data defines the camera's view volume.
      double r = camera.right;
      double b = camera.bottom;
      double t = camera.top;
      double near = - camera.n;

      // Mutate each Vertex object so that it contains normalized camera coordinates.
      for (Vertex v : vertexList)
      {
         double v_x, v_y, v_z;

         if ( camera.perspective )
         {
            // We use two steps to transform the camera's perspective
            // view volume into the normalized perspective view volume.
            // The first step skews the current view volume so that
            // its center line is the negative z-axis. The second step
            // scales the skewed view volume so that it intersects the
            // plane z = -1 with corners (-1, -1, -1) and (+1, +1, -1).

            // For each vertex, skew its coordinates so that the
            // negative z-axis is at the center of the view volume.
            v_z = v.z;
            v_y = v.y - v_z * (t + b)/(2 * near);
            v_x = v.x - v_z * (r + l)/(2 * near);

            // For each vertex, scale its coordinates so that the
            // view volume has corners (-1, -1, -1) and (+1, +1, -1).
            v_x = (2 * near * v_x)/(r - l);
            v_y = (2 * near * v_y)/(t - b);
         }
         else
         {
            // We use two steps to transform the camera's orthographic
            // view volume into the normalized orthographic view volume.
            // The first step translates the current view volume so that
            // its center line is the z-axis. The second step scales the
            // translated view volume so that it intersects the plane
            // z = 0 with corners (-1, -1, 0) and (+1, +1, 0).

            // For each vertex, translate its coordinates so that
            // the z-axis is at the center of the view volume.
            v_x = v.x - (r + l)/2;
            v_y = v.y - (t + b)/2;

            // For each vertex, scale its coordinates so that the
            // view volume has corners (-1, -1, 0) and (+1, +1, 0).
            v_x = (2 * v_x)/(r - l);
            v_y = (2 * v_y)/(t - b);
         }

         // Mutate the Vertex object so that it contains
         // the normalized camera coordinates.
         v.x = v_x;
         v.y = v_y;
      }
   }
}
