/*

*/

package renderer.scene;

import java.util.List;
import java.util.ArrayList;
import java.awt.Color;

/**
   A {@code Model} data structure represents a distinct geometric object
   in a {@link Scene}. A {@code Model} data structure is mainly a {@link List}
   of {@link Vertex} object and another list of {@link LineSegment} objects.
   Each {@link LineSegment} object contains two integers that are the indices
   of two {@link Vertex} objects from the {@code Models}'s vertex list. The
   two {@link Vertex} objects contain the coordinates, in the model's local
   coordinate system, of the line segment's two endpoints.
<p>
   A {@code Model} represent the geometric object as a "wire-frame" of line
   segments, that is, the geometric object is drawn as a collection of "edges".
   This is a fairly simplistic way of doing 3D graphics and we will
   improve this in later renderers.
<p>
   See
<br> <a href="http://en.wikipedia.org/wiki/Wire-frame_model" target="_top">
              http://en.wikipedia.org/wiki/Wire-frame_model</a>
<br>or
<br> <a href="https://www.google.com/search?q=graphics+wireframe&tbm=isch" target="_top">
              https://www.google.com/search?q=graphics+wireframe&tbm=isch</a>
*/
public class Model
{
   public List<Vertex> vertexList = new ArrayList<>();
   public List<LineSegment> lineSegmentList = new ArrayList<>();

   public boolean visible;


   /**
      Construct an empty {@code Model} object.
   */
   public Model()
   {
      visible = true;
   }


   /**
      A "copy constructor". This constructor should make a deep copy
      of the given {@code Model}'s {@link Vertex} list and
      {@link LineSegment} list.

      @param m  {@code Model} to make a copy of
   */
   public Model(Model model) // a "copy constructor"
   {
      super();

      this.visible = model.visible;
      for (Vertex v : model.vertexList)
      {
         this.vertexList.add(new Vertex(v)); // deep copy of each Vertex
      }
      for (LineSegment ls : model.lineSegmentList)
      {
         this.lineSegmentList.add(new LineSegment(ls, this));
      }
   }


   /**
      Add a {@link Vertex} (or vertices) to this {@code Model}'s
      {@link List} of vertices.

      @param vArray  array of {@link Vertex} objects to add to this {@code Model}
   */
   public void addVertex(Vertex... vArray)
   {
      for (Vertex v : vArray)
      {
         this.vertexList.add(new Vertex(v)); // NOTE: deep copy!
      }
   }


   /**
      Add a {@link LineSegment} to this {@code Model}'s {@link List} of
      line segments.
      <p>
      NOTE: This method does not add any vertices to the {@code Model}'s
      {@link Vertex} list. This method assumes that the appropriate vertices
      have been added to the {@code Model}'s {@link Vertex} list.

      @param lsArray  array of {@link LineSegment}s to add to this {@code Model}
   */
   public void addLineSegment(int i0, int i1)
   {
      this.lineSegmentList.add(new LineSegment(i0, i1, this));
   }


   /**
      Add a {@link LineSegment} (or LineSegments) to this {@code Model}'s
      {@link List} of line segments.
      <p>
      This method adds (copies of) the two give {@link Vertex} objects to this
      {@code Model}'s vertex list and then costructs an appropriate
      {@link LineSegment} object to add to this {@code Model}'s line segment list.
      <p>
      This method provides a convenient way to build up a {@code Model} but it
      may not build an efficient {@code Model}. Using this mehtod may lead to
      a lot of redundant {@link Vertex} objects in this {@code Model}'s vertex
      list.

      @param vo  1st {@link Vertex} of {@link LineSegment} to add to this {@code Model}
      @param v1  2nd {@link Vertex} of {@link LineSegment} to add to this {@code Model}
   */
   public void addLineSegment(Vertex v0, Vertex v1)
   {
      int index = this.vertexList.size();
      this.vertexList.add(new Vertex(v0)); // NOTE: deep copy!
      this.vertexList.add(new Vertex(v1)); // NOTE: deep copy!
      this.lineSegmentList.add(new LineSegment(index, index+1, this));
   }


   /**
      Set each {@link Vertex} in this {@code Model}
      to the same {@link Color}.

      @param c  {@link Color} for all of this model's {@link Vertex} objects
   */
   public void setColor(Color c)
   {
      for (Vertex v : this.vertexList)
      {
         v.setColor(c);
      }
   }


   /**
      Set each {@link Vertex} in this {@code Model}
      to the same random {@link Color}.
   */
   public void setColorRandom()
   {
      if ( ! this.vertexList.isEmpty() )
      {
         this.vertexList.get(0).setColorRandom();
         Color c = this.vertexList.get(0).getColor();
         for (Vertex v : this.vertexList)
         {
            v.setColor(c);
         }
      }
   }


   /**
      Set each {@link Vertex} in this {@code Model}
      to a different random {@link Color}.
   */
   public void setRandomColors()
   {
      for (Vertex v : this.vertexList)
      {
         v.setColorRandom();
      }
   }


   /**
      Set each {@link LineSegment} in this {@code Model}
      to a different random {@link Color}.
      <p>
      NOTE: This works best when the {@link LineSegment}
      objects in this {@link Model} do not share any
      {@link Vertex} objects.
   */
   public void setRandomLineSegmentColors()
   {
      for (LineSegment ls : this.lineSegmentList)
      {
         ls.setColorRandom();
      }
   }


   /**
      For debugging.

      @return {@link String} representation of this {@code Model} object
   */
   @Override
   public String toString()
   {
      String result = "";
      result += "Model has " + vertexList.size() + " vertices.\n";
      result += "Model has " + lineSegmentList.size() + " line segments.\n";
      int i = 0;
      for (Vertex v : this.vertexList)
      {
         result += i + ": " + v.toString();
         i++;
      }
      //result = "Printing out this Model's " + lineSegmentList.size() + " Line segments:\n";
      for (LineSegment ls : this.lineSegmentList)
      {
         result += ls.toString();
      }
      //result += "Done printing out Model\n";
      return result;
   }
}
