/*

*/

package renderer.scene;

import java.util.List;
import java.util.ArrayList;

/**
   A {@code Position} data structure represents a geometric object in a
   distinct position (both location and orientation) in three-dimensional
   space as part of a {@link Scene}. A {@code Position} is a recursive
   data structure. Every {@code Position} object contains a {@link List} of
   nested {@code Position} objects. The list of nested {@code Position}s
   lets us define what are called hierarchical, or nested, scenes. These
   are scenes that are made up of groups of models and each group can be
   moved around in the scene as a single unit (while the individual models
   in the group can also be moved around within the group).
<p>
   As a simple example of a group of models in a scene think of a table with
   a lamp and a bowl on it. Once the table, lamp, and bowl are positioned in
   a scene, if we move the table, it is pretty likely that the lamp and bowl
   should move as a group with the table. But we should also be able to
   reposition the bowl, or the lamp, on the table. A scene of a room with
   several tables, chairs grouped around each table, and objects placed on
   each table would be an example of a hierarchical scene with groups of
   models (a table, objects on the table, chairs around the table) that
   contain nested groups of models (each table with its objects on the
   table).
<p>
   As a slightly more complex example of a hierarchical scene, think of a
   solar system with a sun, several planets orbiting around the sun, and
   several moons orbiting around each planet. The solar system as a whole
   is a group of models that we should be able to position in our universe.
   Each planet with its moons is a sub-group of models within the solar
   system. When a planet moves, its moons should move with it, but the
   moons also move around the planet.
<p>
   A {@code Position} object holds a reference to a {@link Model} object,
   a {@link Matrix} object, and a {@code List} of {@code Position} objects.
   The {@link Model} represents the geometric object in the {@link Scene}.
   The {@link Matrix} determines the model's location and orientation in the
   {@link Camera}'s view coordinate system. The {@code Position}'s matrix
   helps us solve the problem of placing and moving a model in a scene.
<p>
   A {@code Position}'s {@code List} of {@code Position} objects creates
   a tree data structure of {@code Position}s. When the renderer renders
   a {@code Position} object, the renderer traverses the tree of
   {@code Position}'s rooted at the given {@code Position}. The renderer
   does a recursive pre-order depth-first-traversal of the tree. As the
   renderer traverses the tree, it accumulates a "current-transformation-matrix"
   that multiplies each Position's matrix along the path from the tree's
   root Position to wherever the traversal is in the tree (this is the
   "pre-order" step in the traversal). The ctm is the current model-to-view
   transformation matrix. The first stage of the rendering pipeline,
   {@link renderer.pipeline.Model2View}, multiplies every {@link Vertex} in
   a {@link Model}'s vertex list by this ctm, which converts the coordinates
   in each {@link Vertex} from the model's own local coordinate system to the
   camera's view coordinate system (which is "shared" by all the other models).
   Multiplication by the ctm has the effect of "placing" the model in view
   space at an appropriate location (using the translation part of the ctm)
   and in the appropriate orientation (using the rotation part of the ctm).
*/
public class Position
{
   public Model  model;
   public Matrix modelMatrix;
   public List<Position> nestedPositions;

   public boolean visible;

   /**
      Construct a default {@code Position} with the identity {@link Matrix},
      no {@link Model} object, and no nested {@code Position}s.
   */
   public Position()
   {
      this.model = null;
      this.modelMatrix = Matrix.identity(); // identity matrix
      this.nestedPositions = new ArrayList<>();
      visible = true;
   }


   /**
      Construct a {@code Position} with the identity {@link Matrix},
      the given {@link Model} object, and no nested {@code Position}s.
   */
   public Position(Model model)
   {
      this.model = model;
      this.modelMatrix = Matrix.identity(); // identity matrix
      this.nestedPositions = new ArrayList<>();
      visible = true;
   }


   /**
      Add a nested {@code Position} (or Positions) to this {@code Position}'s
      {@link List} of nested {@code Position}s.

      @param pArray  array of nested {@code Position}s to add to this {@code Position}
   */
   public void addNestedPosition(Position... pArray)
   {
      for (Position p : pArray)
      {
         this.nestedPositions.add(p);
      }
   }


   /**
      Get the index to a nested {@code Position} from this {@code Position}'s
      {@link List} of nested {@code Position}s.

      @param p  {@code Position} to search for in this {@code Position}'s {@link List} of nested {@code Position}s
      @return the index of the first occurrence of the specified {@code Position},
              or -1 if the {@code Position} list does not contain the {@code Position}
   */
   public int indexOfNestedPosition(Position p)
   {
      return this.nestedPositions.indexOf(p);
   }


   /**
      Get a reference to the nested {@code Position} at the given index in this {@code Position}'s
      {@link List} of nested {@code Position}s.

      @param index  index of the nested {@code Position} to return
      @return nested {@code Position} at the specified index in the {@link List} of nested {@code Position}s
      @throws IndexOutOfBoundsException if the index is out of range
              {@code (index < 0 || index >= size())}
   */
   public Position getNestedPosition(int index)
   {
      return nestedPositions.get(index);
   }


   /**
      Replace the nested {@code Position} at the given index in this {@code Position}'s
      {@link List} of nested {@code Position}s.

      @param index  index of the nested {@code Position} to replace
      @param p      {@code Position} to be stored at the specified index
      @return {@code Position} previously at the specified index
      @throws IndexOutOfBoundsException if the index is out of range
              {@code (index < 0 || index >= size())}
   */
   public Position setNestedPosition(int index, Position p)
   {
      return this.nestedPositions.set(index, p);
   }


   /**
      Reset this {@code Position}'s {@link Matrix} to the identity matrix.

      @return a reference to this {@code Position} to facilitate chaining method calls
   */
   public Position matrix2Identity()
   {
      this.modelMatrix = Matrix.identity();
      return this;
   }

   /**
      Multiply this {@code Position}'s current {@link Matrix} by the
      parameter {@code m} and make the result the new {@link Matrix}.

      @param m  {@link Matrix} to multiply this {@code Position}'s model matrix by
      @return a reference to this {@code Position} to facilitate chaining method calls
   */
   public Position matrixMult(Matrix m)
   {
      this.modelMatrix = this.modelMatrix.times(m);
      return this;
   }

   /**
      Multiply this {@code Position}'s current {@link Matrix} by a rotation
      matrix and make the result the new {@link Matrix}.
      <p>
      The rotation is by angle {@code theta} around the x-axis.

      @param theta  angle (in degrees) to rotate this {@code Position} by
      @return a reference to this {@code Position} to facilitate chaining method calls
   */
   public Position rotateX(double theta)
   {
      this.matrixMult( Matrix.rotateX(theta) );
      return this;
   }

   /**
      Multiply this {@code Position}'s current {@link Matrix} by a rotation
      matrix and make the result the new {@link Matrix}.
      <p>
      The rotation is by angle {@code theta} around the y-axis.

      @param theta  angle (in degrees) to rotate this {@code Position} by
      @return a reference to this {@code Position} to facilitate chaining method calls
   */
   public Position rotateY(double theta)
   {
      this.matrixMult( Matrix.rotateY(theta) );
      return this;
   }

   /**
      Multiply this {@code Position}'s current {@link Matrix} by a rotation
      matrix and make the result the new {@link Matrix}.
      <p>
      The rotation is by angle {@code theta} around the z-axis.

      @param theta  angle (in degrees) to rotate this {@code Position} by
      @return a reference to this {@code Position} to facilitate chaining method calls
   */
   public Position rotateZ(double theta)
   {
      this.matrixMult( Matrix.rotateZ(theta) );
      return this;
   }

   /**
      Multiply this {@code Position}'s current {@link Matrix} by a rotation
      matrix and make the result the new {@link Matrix}.
      <p>
      The rotation is by angle {@code theta} with axis of rotation
      given by the vector {@code (x, y, z)}.

      @param theta  angle (in degrees) to rotate this {@code Position} by
      @param x      x-coordinate of a vector on the axis of rotation
      @param y      y-coordinate of a vector on the axis of rotation
      @param z      z-coordinate of a vector on the axis of rotation
      @return a reference to this {@code Position} to facilitate chaining method calls
   */
   public Position rotate(double theta, double x, double y, double z)
   {
      this.matrixMult( Matrix.rotate(theta, x, y, z) );
      return this;
   }

   /**
      Multiply this {@code Position}'s current {@link Matrix} by a scaling
      matrix and make the result the new {@link Matrix}.
      <p>
      The scaling matrix scales by the amounts {@code x}, {@code y} and
      {@code z} along the x-axis, y-axis, and z-axis (respectively).

      @param x  scaling factor for the x-axis
      @param y  scaling factor for the y-axis
      @param z  scaling factor for the z-axis
      @return a reference to this {@code Position} to facilitate chaining method calls
   */
   public Position scale(double x, double y, double z)
   {
      this.matrixMult( Matrix.scale(x, y, z) );
      return this;
   }

   /**
      Multiply this {@code Position}'s current {@link Matrix} by a scaling
      matrix and make the result the new {@link Matrix}.
      <p>
      The scaling matrix scales by the amount {@code s} along each of
      the x-axis, y-axis, and z-axis.

      @param s  scaling factor for all three directions
      @return a reference to this {@code Position} to facilitate chaining method calls
   */
   public Position scale(double s)
   {
      this.matrixMult( Matrix.scale(s) );
      return this;
   }

   /**
      Multiply this {@code Position}'s current {@link Matrix} by a translation
      matrix and make the result the new {@link Matrix}.
      <p>
      The translation matrix translates by the amounts {@code x}, {@code y},
      and {@code z} along the x-axis, y-axis, and z-axis (respectively).

      @param x  translation amount along the x-axis
      @param y  translation amount along the y-axis
      @param z  translation amount along the z-axis
      @return a reference to this {@code Position} to facilitate chaining method calls
   */
   public Position translate(double x, double y, double z)
   {
      this.matrixMult( Matrix.translate(x, y, z) );
      return this;
   }


   /**
      For debugging.

      @return {@link String} representation of this {@code Position} object
   */
   @Override
   public String toString()
   {
      String result = "";
      result += "The model Matrix is\n";
      result += modelMatrix;
      result += "The Model is\n";
      result += model;
      result += "This Position has " + nestedPositions.size() + " nested Positions\n";
      for (Position p : this.nestedPositions)
      {
         result += p.toString();
      }
      return result;
   }
}
