/*

*/

package renderer.scene;

import java.util.Random;
import java.awt.Color;

/**
   A {@code Vertex} object has four doubles which represent the
   homogeneous coordinates of a point in 3-dimensional space.
   The fourth, homogeneous, coordinate will usually be 1, but in
   some stages of the graphics rendering pipeline it can be some
   other (non-zero) number.
<p>
   When a {@code Vertex} object is created in a client program,
   before the {@code Vertex} object moves down the graphics rendering
   pipeline, the coordinates in the {@code Vertex} will be in some
   model's local coordinate system.
<p>
   As a {@code Vertex} object moves down the graphics rendering
   pipeline, the coordinates in the {@code Vertex} will be transformed
   from one coordinate system to another.
<p>
   A {@code Vertex} object also has a {@link Color} (or "shade"). The
   color of a {@link LineSegment} is interpolated from the colors at
   its two endpoints.
*/
public class Vertex
{
   public double x, y, z, w;   // a vertex in homogenous coordinates
   public float  r = 1.0f, g = 1.0f, b = 1.0f;  // color at this vertex


   /**
      Construct a default {@code Vertex}.
   */
   public Vertex()
   {
      set(0.0, 0.0, 0.0, 1.0);
   }


   /**
      Construct a new {@code Vertex} (with homogeneous coordinates)
      using the given {@code x}, {@code y}, and {@code z} coordinates.

      @param x  x-coordinate of the new {@code Vertex}
      @param y  y-coordinate of the new {@code Vertex}
      @param z  z-coordinate of the new {@code Vertex}
   */
   public Vertex(double x, double y, double z)
   {
      set(x, y, z, 1.0);
   }


   /**
      Construct a new {@code Vertex} (with homogeneous coordinates)
      using the given {@code x}, {@code y}, and {@code z} coordinates
      and the given {@link Color}.

      @param x  x-coordinate of the new {@code Vertex}
      @param y  y-coordinate of the new {@code Vertex}
      @param z  z-coordinate of the new {@code Vertex}
      @param c  {@link Color} of the new {@code Vertex}
   */
   public Vertex(double x, double y, double z, Color c)
   {
      set(x, y, z, 1.0);
      setColor(c);
   }


   /**
      Construct a new {@code Vertex} with the given homogeneous coordinates.

      @param x  x-coordinate of the new {@code Vertex}
      @param y  y-coordinate of the new {@code Vertex}
      @param z  z-coordinate of the new {@code Vertex}
      @param w  w-coordinate of the new {@code Vertex}
   */
   public Vertex(double x, double y, double z, double w)
   {
      set(x, y, z, w);
   }


   /**
      Construct a new {@code Vertex} with the given homogeneous coordinates
      and the given {@link Color}.

      @param x  x-coordinate of the new {@code Vertex}
      @param y  y-coordinate of the new {@code Vertex}
      @param z  z-coordinate of the new {@code Vertex}
      @param w  w-coordinate of the new {@code Vertex}
      @param c  {@link Color} of the new {@code Vertex}
   */
   public Vertex(double x, double y, double z, double w, Color c)
   {
      set(x, y, z, w);
      setColor(c);
   }


   /**
      Construct a new {@code Vertex} that is a copy of another {@code Vertex}.

      @param v  {@code Vertex} to make a copy of
   */
   public Vertex(Vertex v) // a "copy constructor"
   {
      set(v.x, v.y, v.z, v.w);
      setColor(v);
   }


   /**
      Set the {@code x}, {@code y}, and {@code z} coordinates
      of this {@code Vertex}.

      @param x  new x-coordinate for this {@code Vertex}
      @param y  new y-coordinate for this {@code Vertex}
      @param z  new z-coordinate for this {@code Vertex}
   */
   public void set(double x, double y, double z)
   {
      set(x, y, z, 1);
   }


   /**
      Set the homogeneous coordinates of this {@code Vertex}.

      @param x  new x-coordinate for this {@code Vertex}
      @param y  new y-coordinate for this {@code Vertex}
      @param z  new z-coordinate for this {@code Vertex}
      @param w  new w-coordinate for this {@code Vertex}
   */
   public void set(double x, double y, double z, double w)
   {
      this.x = x;
      this.y = y;
      this.z = z;
      this.w = w;
   }


   /**
      {@link Matrix} times {@code Vertex} mutates this {@code Vertex}.
      <p>
      Needed by the linear transformation stages of the graphics pipeline.

      @param matrix  {@code Matrix} to multiply this {@code Vertex} by
   */
   public void times(Matrix matrix)
   {
      Vector col1 = matrix.v1;
      Vector col2 = matrix.v2;
      Vector col3 = matrix.v3;
      Vector col4 = matrix.v4;

      // dot product of the first matrix row with this vertex
      double x2 = (col1.x * this.x) + (col2.x * this.y)  + (col3.x * this.z) + (col4.x * this.w);

      // dot product of the second matrix row with this vertex
      double y2 = (col1.y * this.x) + (col2.y * this.y)  + (col3.y * this.z) + (col4.y * this.w);

      // dot product of the third matrix row with this vertex
      double z2 = (col1.z * this.x) + (col2.z * this.y)  + (col3.z * this.z) + (col4.z * this.w);

      // dot product of the fourth matrix row with this vertex
      double w2 = (col1.w * this.x) + (col2.w * this.y)  + (col3.w * this.z) + (col4.w * this.w);

      // Now mutate this Vertex object.
      this.x = x2;
      this.y = y2;
      this.z = z2;
      this.w = w2;
   }


   /**
      Get the {@link Color} of this {@code Vertex}.

      @return the {@link Color} of this {@code Vertex} object
   */
   public Color getColor()
   {
      return new Color(this.r, this.g, this.b);
   }


   /**
      Set the {@link Color} of this {@code Vertex}
      using floats between 0 and 1.

      @param r  red color value for this {@code Vertex} as a float between 0 and 1
      @param g  green color value for this {@code Vertex} as a float between 0 and 1
      @param b  blue color value for this {@code Vertex} as a float between 0 and 1
   */
   public void setColor(float r, float g, float b)
   {
      this.r = r;
      this.g = g;
      this.b = b;

      if ( (r < 0.0) || (r > 1.0)
        || (g < 0.0) || (g > 1.0)
        || (b < 0.0) || (b > 1.0) )
      {
         System.err.println("ERROR! Invalid float color for vertex");
         System.err.print( this.toString() );
         System.err.printf("<r,g,b> = <% .5f  % .5f  % .5f>\n", r, g, b);
         Thread.dumpStack();
       //System.err.println(Arrays.toString(Thread.currentThread().getStackTrace()));
         System.exit(-1);
      }
   }


   /**
      Set the {@link Color} of this {@code Vertex}
      using ints between 0 and 255.

      @param r  red color value for this {@code Vertex} as an integer between 0 and 255
      @param g  green color value for this {@code Vertex} as an integer between 0 and 255
      @param b  blue color value for this {@code Vertex} as an integer between 0 and 255
   */
   public void setColor(int r, int g, int b)
   {
      this.r = ((float)r)/(float)255;
      this.g = ((float)g)/(float)255;
      this.b = ((float)b)/(float)255;

      if ( (r < 0) || (r > 255)
        || (g < 0) || (g > 255)
        || (b < 0) || (b > 255))
      {
         System.err.println("ERROR! Invalid int color for vertex");
         System.err.print( this.toString() );
         System.err.printf("<r,g,b> = <%d  %d  %d>\n", r, g, b);
         Thread.dumpStack();
       //System.err.println(Arrays.toString(Thread.currentThread().getStackTrace()));
         System.exit(-1);
      }
   }


   /**
      Set the {@link Color} of this {@code Vertex}
      using a {@link Color} object.

      @param c  {@link Color} for this {@code Vertex} object
   */
   public void setColor(Color c)
   {
      setColor(c.getRed(), c.getGreen(), c.getBlue());
   }


   /**
      Set the {@link Color} of this {@code Vertex}
      using the colors from another {@code Vertex}.

      @param v  {@code Vertex} object to get color values from
   */
   public void setColor(Vertex v)
   {
      // copy the color from v to this vertex
      setColor(v.r, v.g, v.b);
   }


   /**
      Set the {@link Color} of this {@code Vertex} to a random color.
   */
   public void setColorRandom()
   {
      Random generator = new Random();
      float r = generator.nextFloat();
      float g = generator.nextFloat();
      float b = generator.nextFloat();
      setColor(r, g, b);
   }


   /**
      For debugging.

      @return {@link String} representation of this {@code Vertex} object
   */
   @Override
   public String toString()
   {
      int precision = 5;  // the default precision for the format string
      return toString(precision);
   }


   /**
      For debugging.
      <p>
      Allow the precision of the formatted output to be specified.

      @param precision  precision value for the format string
      @return {@link String} representation of this {@code Vertex} object
   */
   public String toString(int precision)
   {
      int iWidth = 3; // default width of integer part of the format string
      return toString(precision, iWidth);
   }


   /**
      For debugging.
      <p>
      Allow the precision and width of the formatted output to be specified.
      By width, we mean the width of the integer part of each number.

      @param precision  precision value for the format string
      @param iWidth     width of the integer part of the format string
      @return {@link String} representation of this {@code Vertex} object
   */
   public String toString(int precision, int iWidth)
   {
      // Here is one way to get programmable precision and width.
      int p = precision;      // the precision for the following format string
      int t = p + iWidth + 2; // the width for the following format string
      String format = "(x,y,z,w)=(% "+t+"."+p+"f  % "+t+"."+p+"f  % "+t+"."+p+"f  % "+t+"."+p+"f)\n";
      return String.format(format, x, y, z, w);
   }


   /**
      For debugging.

      @return {@link String} representation of the color information in this {@code Vertex}
   */
   public String toStringColor()
   {
      return String.format("{r,g,b} = {% 10.5f  % 10.5f  % 10.5f}\n", r, g, b);
   }
}
